@@ -4,16 +4,20 @@
 #ifndef CORBA_SMALL_H
 #define CORBA_SMALL_H 1
 
-#include <orbit/GIOP/giop.h>
+#include <glib-object.h>
 #include <orbit/orb-core/orbit-interface.h>
 
+G_BEGIN_DECLS
+
 typedef struct {
 	CORBA_unsigned_long           version;
 	ORBit_IInterface            **interfaces;
 	CORBA_sequence_CORBA_TypeCode types;
 } ORBit_IModule;
 
+const char    *ORBit_get_safe_tmp      (void);
 
+/* Builtin allocators */
 gpointer       ORBit_small_alloc       (CORBA_TypeCode      tc);
 gpointer       ORBit_small_allocbuf    (CORBA_TypeCode      tc,
 					CORBA_unsigned_long length);
@@ -21,6 +25,17 @@ void           ORBit_small_freekids    (CORBA_TypeCode      tc,
 					gpointer            p,
 					gpointer            d);
 
+/* More friendly(?) sequence allocators */
+gpointer       ORBit_sequence_alloc    (CORBA_TypeCode      sequence_tc,
+					CORBA_unsigned_long length);
+void           ORBit_sequence_append   (gpointer            sequence,
+					gconstpointer       element);
+void           ORBit_sequence_set_size (gpointer            sequence,
+					CORBA_unsigned_long length);
+#define        ORBit_sequence_index(sequence,idx) (sequence)->_buffer[(idx)]
+void           ORBit_sequence_concat   (gpointer            sequence,
+					gconstpointer       append);
+
 typedef enum {
 	ORBIT_CONNECTION_CONNECTED,
 	ORBIT_CONNECTION_CONNECTING,
@@ -28,22 +43,34 @@ typedef enum {
 	ORBIT_CONNECTION_IN_PROC
 } ORBitConnectionStatus;
 
-gpointer              ORBit_small_get_servant           (CORBA_Object obj);
-ORBitConnectionStatus ORBit_small_get_connection_status (CORBA_Object obj);
-ORBitConnectionStatus ORBit_small_listen_for_broken     (CORBA_Object obj,
-							 GCallback    fn,
-							 gpointer     user_data);
-ORBitConnectionStatus ORBit_small_unlisten_for_broken   (CORBA_Object obj,
-							 GCallback    fn);
+gpointer              ORBit_small_get_servant              (CORBA_Object obj);
+ORBitConnectionStatus ORBit_small_get_connection_status    (CORBA_Object obj);
+ORBitConnectionStatus ORBit_small_listen_for_broken        (CORBA_Object obj,
+							    GCallback    fn,
+							    gpointer     user_data);
+ORBitConnectionStatus ORBit_small_unlisten_for_broken_full (CORBA_Object obj,
+							    GCallback    fn,
+							    gpointer     user_data);
+/* old / stale */
+ORBitConnectionStatus ORBit_small_unlisten_for_broken      (CORBA_Object obj,
+							    GCallback    fn);
 
 typedef struct _ORBitConnection ORBitConnection;
 
 ORBitConnection      *ORBit_small_get_connection        (CORBA_Object     obj);
+ORBitConnection      *ORBit_small_get_connection_ref    (CORBA_Object     obj);
+void                  ORBit_small_connection_unref      (ORBitConnection *cnx);
 void                  ORBit_connection_set_max_buffer   (ORBitConnection *cnx,
 							 gulong           max_buffer_bytes);
 
 #if defined(ORBIT2_INTERNAL_API) || defined (ORBIT2_STUBS_API)
 
+#define ORBIT_SMALL_FAST_LOCALS           1
+#define ORBIT_SMALL_FORCE_GENERIC_MARSHAL 2
+
+extern int     ORBit_small_flags;
+
+/* Deprecated - only for bin-compat with pre 2.4 stubs */
 void           ORBit_small_invoke_stub (CORBA_Object        object,
 					ORBit_IMethod      *m_data,
 					gpointer            ret,
@@ -51,10 +78,21 @@ void           ORBit_small_invoke_stub (CORBA_Object        object,
 					CORBA_Context       ctx,
 					CORBA_Environment  *ev);
 
+
+void           ORBit_small_invoke_stub_n (CORBA_Object        object,
+					  ORBit_IMethods     *methods,
+					  glong               index,
+					  gpointer            ret,
+					  gpointer           *args,
+					  CORBA_Context       ctx,
+					  CORBA_Environment  *ev);
+
 #endif /* defined(ORBIT2_INTERNAL_API) || defined (ORBIT2_STUBS_API) */
 
 #ifdef ORBIT2_INTERNAL_API
 
+#include <orbit/GIOP/giop.h>
+
 void           ORBit_small_invoke_adaptor (ORBit_OAObject     adaptor_obj,
 					   GIOPRecvBuffer    *recv_buffer,
 					   ORBit_IMethod     *m_data,
@@ -97,4 +135,6 @@ void ORBit_small_demarshal_async     (ORBitAsyncQueueEntry  *aqe,
 				      gpointer              *args,
 				      CORBA_Environment     *ev);
 
+G_END_DECLS
+
 #endif /* CORBA_SMALL_H */
