@@ -9,13 +9,13 @@ free_entry (gpointer key, gpointer value, gpointer user_data)
 {
 	g_free (key);
 	g_free (value);
-	
+
 	return TRUE;
 }
 
 static void ORBit_Context_free_fn(ORBit_RootObject ctx);
 
-static gboolean
+static void
 free_child (gpointer value, gpointer user_data)
 {
 	CORBA_Context ctx = value;
@@ -23,8 +23,6 @@ free_child (gpointer value, gpointer user_data)
 	ctx->parent.refs = 0;
 	ctx->parent_ctx = CORBA_OBJECT_NIL;
 	ORBit_Context_free_fn (ORBIT_ROOT_OBJECT (ctx));
-
-	return TRUE;
 }
 
 static void
@@ -33,18 +31,18 @@ ORBit_Context_free_fn (ORBit_RootObject obj_in)
 	CORBA_Context ctx = (CORBA_Context) obj_in;
 
 	if (ctx->children) {
-		g_slist_foreach (ctx->children, (GFunc)free_child, ctx);
+		g_slist_foreach (ctx->children, free_child, NULL);
 		g_slist_free (ctx->children);
 	}
 
 	if (ctx->mappings) {
-		g_hash_table_foreach_remove (ctx->mappings, free_entry, ctx);
+		g_hash_table_foreach_remove (ctx->mappings, free_entry, NULL);
 		g_hash_table_destroy (ctx->mappings);
 	}
 
 	if (ctx->parent_ctx != CORBA_OBJECT_NIL)
 		ctx->parent_ctx->children = g_slist_remove (
-			ctx->parent_ctx->children, ctx->the_name);
+			ctx->parent_ctx->children, ctx);
 	  
 	g_free (ctx->the_name);
 	
