@@ -36,7 +36,7 @@
 #include "orb-core-private.h"
 #include "orbit-debug.h"
 
-int ORBit_small_flags = ORBIT_SMALL_FAST_LOCALS;
+int ORBit_small_flags = 0;
 
 gpointer
 ORBit_small_alloc (CORBA_TypeCode tc)
@@ -322,7 +322,8 @@ orbit_small_marshal (CORBA_Object           obj,
 	}
 
 	send_buffer = giop_send_buffer_use_request (
-		cnx->giop_version, request_id, CORBA_TRUE,
+		cnx->giop_version, request_id, 
+		(m_data->flags & ORBit_I_METHOD_1_WAY) == 0,
 		obj->object_key, &op_vec, NULL);
 
 	if (!send_buffer)
@@ -381,27 +382,19 @@ typedef enum {
 	MARSHAL_CLEAN
 } DeMarshalRetType;
 
-typedef enum {
-	RECV_BUFFER_RELEASE,
-	RECV_BUFFER_KEEP
-} DeMarshalRecvRelease;
-
 static DeMarshalRetType
 orbit_small_demarshal (CORBA_Object           obj,
 		       GIOPConnection       **cnx,
-		       GIOPMessageQueueEntry *mqe,
+		       GIOPRecvBuffer        *recv_buffer,
 		       CORBA_Environment     *ev,
 		       gpointer               ret,
 		       ORBit_IMethod         *m_data,
-		       gpointer              *args,
-		       DeMarshalRecvRelease   recv_release)
+		       gpointer              *args)
 {
 	gpointer        data, p;
 	CORBA_TypeCode  tc;
-	GIOPRecvBuffer *recv_buffer;
 	CORBA_ORB       orb = obj->orb;
 
-	recv_buffer = giop_recv_buffer_get (mqe);
 	if (!recv_buffer) {
 		dprintf (MESSAGES, "No recv buffer ...\n");
 		return MARSHAL_SYS_EXCEPTION_INCOMPLETE;
@@ -471,6 +464,8 @@ orbit_small_demarshal (CORBA_Object           obj,
 			dump_arg (a, tc);
 
 			if (a->flags & ORBit_I_ARG_OUT)
+				/* this may read (&discard) uninitialized memory,
+				 * see 'foo' below. This is for simplicity. */
 				arg = *(gpointer *)args [i];
 			else
 				arg = args [i];
@@ -505,7 +500,7 @@ orbit_small_demarshal (CORBA_Object           obj,
 				} else if (a->flags & ORBit_I_ARG_INOUT) {
 					ORBit_freekids_via_TypeCode (tc, arg);
 					do_demarshal_value (recv_buffer, &arg, tc, orb);
-				} else
+				} else /* 'foo' - don't use the bogus 'arg' contents */
 					*(gpointer *)args [i] = p = ORBit_demarshal_arg (
 						recv_buffer, tc, obj->orb);
 
@@ -522,9 +517,6 @@ orbit_small_demarshal (CORBA_Object           obj,
 			tprintf (" )");
 	}
 
-	if (recv_release == RECV_BUFFER_RELEASE)
-		giop_recv_buffer_unuse (recv_buffer);
-
 	return MARSHAL_CLEAN;
 
  demarshal_exception:
@@ -540,17 +532,11 @@ orbit_small_demarshal (CORBA_Object           obj,
 		*cnx = ORBit_handle_location_forward (recv_buffer, obj);
 		tprintf (" Exception: forward");
 
-		if (recv_release == RECV_BUFFER_RELEASE)
-			giop_recv_buffer_unuse (recv_buffer);
-
 		return MARSHAL_RETRY;
 	} else {
 		ORBit_handle_exception_array (
 			recv_buffer, ev, &m_data->exceptions, obj->orb);
 
-		if (recv_release == RECV_BUFFER_RELEASE)
-			giop_recv_buffer_unuse (recv_buffer);
-
 #ifdef G_ENABLE_DEBUG
 		if (_orbit_debug_flags & ORBIT_DEBUG_TRACES) {
 			if (ev->_major == CORBA_SYSTEM_EXCEPTION)
@@ -575,8 +561,14 @@ ORBit_small_invoke_stub_n (CORBA_Object        object,
 			   CORBA_Context       ctx,
 			   CORBA_Environment  *ev)
 {
-	/* FIXME: validate on the sequence length ? */
-	ORBit_small_invoke_stub (object, &methods->_buffer[index], ret, args, ctx, ev);
+	if (index < 0 || index > methods->_length) {
+		dprintf (MESSAGES, "Cannot invoke OOB method (%ld,%lu)\n",
+			 index, (gulong)methods->_length);
+		CORBA_exception_set_system (ev, ex_CORBA_NO_IMPLEMENT,
+					    CORBA_COMPLETED_NO);
+
+	} else
+		ORBit_small_invoke_stub (object, &methods->_buffer[index], ret, args, ctx, ev);
 }
 
 void
@@ -589,9 +581,11 @@ ORBit_small_invoke_stub (CORBA_Object       obj,
 {
 	CORBA_unsigned_long     request_id;
 	CORBA_completion_status completion_status;
-	GIOPConnection         *cnx;
+	GIOPConnection         *cnx = NULL;
 	GIOPMessageQueueEntry   mqe;
 	ORBit_OAObject          adaptor_obj;
+	GIOPRecvBuffer         *recv_buffer = NULL;
+	CORBA_Object            xt_proxy = CORBA_OBJECT_NIL;
 
 	if (!obj) {
 		dprintf (MESSAGES, "Cannot invoke method on null object\n");
@@ -603,12 +597,20 @@ ORBit_small_invoke_stub (CORBA_Object       obj,
 	adaptor_obj = obj->adaptor_obj;
 
 	if (adaptor_obj) {
-		tprintf_header (obj, m_data);
-		tprintf ("[in-proc]");
-		ORBit_small_handle_request (adaptor_obj, m_data->name, ret,
-					    args, ctx, NULL, ev);
-		goto clean_out;
-	}
+		/* FIXME: unchecked cast */
+		if (ORBit_poa_allow_cross_thread_call ((ORBit_POAObject) adaptor_obj)) {
+			tprintf_header (obj, m_data);
+			tprintf ("[in-proc]");
+			ORBit_small_handle_request (adaptor_obj, m_data->name, ret,
+						    args, ctx, NULL, ev);
+			goto clean_out;
+		} else {
+			tprintf ("[in-proc-XT]");
+			xt_proxy = ORBit_objref_get_proxy (obj);
+			obj = xt_proxy;
+		}
+	} else
+		giop_thread_new_check (NULL);
 
 	cnx = ORBit_object_get_connection (obj);
 
@@ -636,9 +638,11 @@ ORBit_small_invoke_stub (CORBA_Object       obj,
 		goto clean_out;
 	}
 
-	switch (orbit_small_demarshal (obj, &cnx, &mqe, ev,
-				       ret, m_data, args,
-				       RECV_BUFFER_RELEASE))
+	giop_recv_buffer_unuse (recv_buffer);
+	recv_buffer = giop_recv_buffer_get (&mqe);
+
+	switch (orbit_small_demarshal (obj, &cnx, recv_buffer, ev,
+				       ret, m_data, args))
 	{
 	case MARSHAL_SYS_EXCEPTION_COMPLETE:
 		completion_status = CORBA_COMPLETED_YES;
@@ -663,7 +667,12 @@ ORBit_small_invoke_stub (CORBA_Object       obj,
 	};
 
  clean_out:
+	ORBit_RootObject_release (xt_proxy);
+	giop_recv_buffer_unuse (recv_buffer);
+
 	tprintf_end_method ();
+	if (cnx)
+		giop_connection_unref (cnx);
 	return;
 
  system_exception:
@@ -1044,8 +1053,10 @@ ORBit_small_demarshal_async (ORBitAsyncQueueEntry *aqe,
 			     gpointer             *args,
 			     CORBA_Environment    *ev)
 {
-	switch (orbit_small_demarshal (aqe->obj, &aqe->mqe.cnx, &aqe->mqe, ev,
-				       ret, aqe->m_data, args, RECV_BUFFER_KEEP)) {
+	g_return_if_fail (aqe->mqe.buffer != NULL);
+
+	switch (orbit_small_demarshal (aqe->obj, &aqe->mqe.cnx, aqe->mqe.buffer, ev,
+				       ret, aqe->m_data, args)) {
 	case MARSHAL_SYS_EXCEPTION_COMPLETE:
 		aqe->completion_status = CORBA_COMPLETED_YES;
 		dprintf (MESSAGES, "Sys exception completed on id 0x%x\n\n",
@@ -1094,7 +1105,7 @@ async_recv_cb (ORBitAsyncQueueEntry *aqe)
 	aqe->mqe.async_cb = NULL;
 
 	if (!aqe->mqe.cnx ||
-	    aqe->mqe.cnx->parent.status == LINC_DISCONNECTED)
+	    aqe->mqe.cnx->parent.status == LINK_DISCONNECTED)
 		CORBA_exception_set_system (ev, ex_CORBA_COMM_FAILURE,
 					    aqe->completion_status);
 
@@ -1172,6 +1183,8 @@ ORBit_small_invoke_async (CORBA_Object         obj,
 	aqe->m_data = /* ORBit_RootObject_duplicate */ (m_data);
 
  clean_out:
+	if (cnx)
+		giop_connection_unref (cnx);
 	tprintf_end_method ();
 	return;
 
@@ -1202,17 +1215,17 @@ ORBit_small_get_servant (CORBA_Object obj)
 }
 
 static ORBitConnectionStatus
-get_status (LINCConnection *cnx)
+get_status (GIOPConnection *cnx)
 {
 	ORBitConnectionStatus ret;
 
 	g_return_val_if_fail (cnx != NULL, ORBIT_CONNECTION_DISCONNECTED);
 
-	switch (cnx->status) {
-	case LINC_CONNECTED:
+	switch (link_connection_get_status (LINK_CONNECTION (cnx))) {
+	case LINK_CONNECTED:
 		ret = ORBIT_CONNECTION_CONNECTED;
 		break;
-	case LINC_CONNECTING:
+	case LINK_CONNECTING:
 		ret = ORBIT_CONNECTION_CONNECTED;
 		break;
 	default:
@@ -1227,16 +1240,20 @@ ORBit_small_get_connection_status (CORBA_Object obj)
 {
 	ORBitConnectionStatus ret;
 
+	g_return_val_if_fail (obj != CORBA_OBJECT_NIL,
+			      ORBIT_CONNECTION_DISCONNECTED);
+
 	if (ORBit_small_get_servant (obj))
 		ret = ORBIT_CONNECTION_IN_PROC;
 	else {
-		GIOPConnection *connection;
+		GIOPConnection *cnx;
 
-		connection = ORBit_object_get_connection (obj);
+		cnx = ORBit_object_get_connection (obj);
 
-		if (connection)
-			ret = get_status (LINC_CONNECTION (connection));
-		else
+		if (cnx) {
+			ret = get_status (cnx);
+			giop_connection_unref (cnx);
+		} else
 			ret = ORBIT_CONNECTION_DISCONNECTED;
 	}
 
@@ -1257,14 +1274,52 @@ ORBit_small_listen_for_broken (CORBA_Object obj,
 		ret = ORBIT_CONNECTION_IN_PROC;
 
 	else {
-		GIOPConnection *connection;
+		GIOPConnection *cnx;
 
-		connection = ORBit_object_get_connection (obj);
+		cnx = ORBit_object_get_connection (obj);
 
-		if (connection) {
-			ret = get_status (LINC_CONNECTION (connection));
-			g_signal_connect (G_OBJECT (connection),
-					  "broken", fn, user_data);
+		if (cnx) {
+			ret = get_status (cnx);
+			g_signal_connect (cnx, "broken", fn, user_data);
+			giop_connection_unref (cnx);
+		} else
+			ret = ORBIT_CONNECTION_DISCONNECTED;
+	}
+
+	return ret;
+}
+
+ORBitConnectionStatus
+ORBit_small_unlisten_for_broken_full (CORBA_Object obj,
+				      GCallback    fn,
+				      gpointer     user_data)
+{
+	ORBitConnectionStatus ret;
+
+	if (!obj)
+		ret = ORBIT_CONNECTION_DISCONNECTED;
+
+	else if (ORBit_small_get_servant (obj))
+		ret = ORBIT_CONNECTION_IN_PROC;
+
+	else {
+		GIOPConnection *cnx;
+
+		cnx = ORBit_object_get_connection (obj);
+
+		if (cnx) {
+			GSignalMatchType t = 0;
+
+			ret = get_status (cnx);
+			if (fn)
+				t |= G_SIGNAL_MATCH_FUNC;
+			if (user_data)
+				t |= G_SIGNAL_MATCH_DATA;
+
+			g_signal_handlers_disconnect_matched
+				(cnx, t, 0, 0, NULL,
+				 G_CALLBACK (fn), user_data);
+			giop_connection_unref (cnx);
 		} else
 			ret = ORBIT_CONNECTION_DISCONNECTED;
 	}
@@ -1276,46 +1331,42 @@ ORBitConnectionStatus
 ORBit_small_unlisten_for_broken (CORBA_Object obj,
 				 GCallback    fn)
 {
-	ORBitConnectionStatus ret;
+	return ORBit_small_unlisten_for_broken_full (obj, fn, NULL);
+}
 
-	if (!obj)
-		ret = ORBIT_CONNECTION_DISCONNECTED;
-
-	else if (ORBit_small_get_servant (obj))
-		ret = ORBIT_CONNECTION_IN_PROC;
-
-	else {
-		GIOPConnection *connection;
-
-		connection = ORBit_object_get_connection (obj);
-
-		if (connection) {
-			ret = get_status (LINC_CONNECTION (connection));
-			g_signal_handlers_disconnect_matched (
-				G_OBJECT (connection), 
-				G_SIGNAL_MATCH_FUNC,
-				0, 0, NULL, G_CALLBACK (fn), NULL);
-		} else
-			ret = ORBIT_CONNECTION_DISCONNECTED;
-	}
-
-	return ret;
+ORBitConnection *
+ORBit_small_get_connection_ref (CORBA_Object obj)
+{
+	return (ORBitConnection *) ORBit_object_get_connection (obj);
 }
 
 ORBitConnection *
 ORBit_small_get_connection (CORBA_Object obj)
 {
-	return (ORBitConnection *) ORBit_object_get_connection (obj);
+	ORBitConnection *cnx;
+
+	cnx = ORBit_small_get_connection_ref (obj);
+
+	/* This sucks but compatibily */
+	ORBit_small_connection_unref (cnx);
+	
+	return cnx;
+}
+
+void
+ORBit_small_connection_unref (ORBitConnection *cnx)
+{
+	if (cnx)
+		giop_connection_unref (cnx);
 }
 
 void
 ORBit_connection_set_max_buffer (ORBitConnection *cnx,
 				 gulong           max_buffer_bytes)
 {
-	LINCConnection *lcnx = (LINCConnection *) cnx;
+	LinkConnection *lcnx = (LinkConnection *) cnx;
 
-	g_return_if_fail (LINC_IS_CONNECTION (lcnx));
+	g_return_if_fail (LINK_IS_CONNECTION (lcnx));
 
-	linc_connection_set_max_buffer (lcnx, max_buffer_bytes);
+	link_connection_set_max_buffer (lcnx, max_buffer_bytes);
 }
-
