@@ -1,6 +1,7 @@
 #include <string.h>
 #include <stdlib.h>
 #include <unistd.h>
+#include <stdarg.h>
 
 #include <orbit/orbit.h>
 #include <orb-core-export.h>
@@ -14,8 +15,16 @@
 #include "poa-private.h"
 #include "orbit-poa.h"
 
-#define POA_LOCK(poa)   LINC_MUTEX_LOCK(poa->base.lock)
-#define POA_UNLOCK(poa) LINC_MUTEX_UNLOCK(poa->base.lock)
+#define ORBIT_SERVANT_SET_CLASSINFO(servant,ci) {                      \
+  ((PortableServer_ServantBase *)(servant))->vepv[0]->_private = (ci); \
+}
+
+#define POA_LOCK(poa)   LINK_MUTEX_LOCK(poa->base.lock)
+#define POA_UNLOCK(poa) LINK_MUTEX_UNLOCK(poa->base.lock)
+
+static GMutex     *ORBit_class_assignment_lock    = NULL;
+static GHashTable *ORBit_class_assignments        = NULL;
+static guint       ORBit_class_assignment_counter = 0;
 
 static PortableServer_Servant ORBit_POA_ServantManager_use_servant(
 				     PortableServer_POA poa,
@@ -58,10 +67,7 @@ static void               ORBit_POA_deactivate_object       (PortableServer_POA
 							     ORBit_POAObject     pobj,
 							     CORBA_boolean       do_etherealize,
 							     CORBA_boolean       is_cleanup);
-
-static GHashTable *ORBit_class_assignments = NULL;
-static guint ORBit_class_assignment_counter = 0;
-
+     
 /* PortableServer_Current interface */
 static void
 ORBit_POACurrent_free_fn (ORBit_RootObject obj_in)
@@ -102,14 +108,14 @@ ORBit_POACurrent_get_object (PortableServer_Current  obj,
 
 	g_assert (obj && obj->parent.interface->type == ORBIT_ROT_POACURRENT);
 
-	LINC_MUTEX_LOCK (obj->orb->lock);
+	LINK_MUTEX_LOCK (obj->orb->lock);
 	if (obj->orb->current_invocations == NULL)
 		CORBA_exception_set_system
 			(ev, ex_PortableServer_Current_NoContext,
 			 CORBA_COMPLETED_NO);
 	else
 		pobj = (ORBit_POAObject) obj->orb->current_invocations->data;
-	LINC_MUTEX_UNLOCK (obj->orb->lock);
+	LINK_MUTEX_UNLOCK (obj->orb->lock);
 
 	return pobj;
 }
@@ -117,12 +123,17 @@ ORBit_POACurrent_get_object (PortableServer_Current  obj,
 PortableServer_ClassInfo *
 ORBit_classinfo_lookup (const char *type_id)
 {
-	if (!ORBit_class_assignments)
-		return NULL;
+	PortableServer_ClassInfo *ci = NULL;
 
-	return g_hash_table_lookup (ORBit_class_assignments, type_id);
+	LINK_MUTEX_LOCK   (ORBit_class_assignment_lock);
+	if (ORBit_class_assignments)
+		ci = g_hash_table_lookup (ORBit_class_assignments, type_id);
+	LINK_MUTEX_UNLOCK (ORBit_class_assignment_lock);
+
+	return ci;
 }
 
+/* Deprecated & not thread-safe */
 void
 ORBit_classinfo_register (PortableServer_ClassInfo *ci)
 {
@@ -143,6 +154,57 @@ ORBit_classinfo_register (PortableServer_ClassInfo *ci)
 			     (gpointer) ci->class_name, ci);
 }
 
+void
+ORBit_skel_class_register (PortableServer_ClassInfo   *ci,
+			   PortableServer_ServantBase *servant,
+			   void                      (*opt_finalize) (PortableServer_Servant,
+								      CORBA_Environment *),
+			   CORBA_unsigned_long         class_offset,
+			   CORBA_unsigned_long         first_parent_id,
+			   ...)
+{
+	va_list args;
+	CORBA_unsigned_long id;
+
+	va_start (args, first_parent_id);
+
+	/*
+	 * FIXME: a double check/lock/write barrier pattern
+	 * would be far faster here.
+	 */
+	LINK_MUTEX_LOCK (ORBit_class_assignment_lock);
+
+	ORBit_classinfo_register (ci);
+	if (!ci->vepvmap) {
+		CORBA_unsigned_long offset;
+
+		/* FIXME: we could propagate the size nicely here */
+		ci->vepvmap = g_new0 (ORBit_VepvIdx, 
+				      *(ci->class_id) + 1);
+
+		ci->vepvmap[*(ci->class_id)] = class_offset;
+		
+		for (id = first_parent_id; id;) {
+
+			offset = va_arg (args, CORBA_unsigned_long);
+			
+			g_assert (id <= *(ci->class_id));
+			ci->vepvmap [id] = offset;
+
+			id = va_arg (args, CORBA_unsigned_long);
+		}
+	}
+	
+	LINK_MUTEX_UNLOCK (ORBit_class_assignment_lock);
+
+	if (!servant->vepv[0]->finalize)
+		servant->vepv[0]->finalize = opt_finalize;
+
+	ORBIT_SERVANT_SET_CLASSINFO (servant, ci);
+
+	va_end (args);
+}
+
 static void
 check_child_poa_inuse_T (char               *name,
 			 PortableServer_POA  poa,
@@ -463,6 +525,8 @@ ORBit_POA_free_fn (ORBit_RootObject obj)
 	ORBit_ObjectAdaptor adaptor = (ORBit_ObjectAdaptor) obj;
 	PortableServer_POA  poa = (PortableServer_POA) obj;
 
+	giop_thread_key_release (obj);
+
 	if (adaptor->adaptor_key._buffer)
 		ORBit_free_T (adaptor->adaptor_key._buffer);
 
@@ -561,6 +625,19 @@ ORBit_POA_set_policy (PortableServer_POA  poa,
 	}
 }
 
+static void
+ORBit_POA_copy_policies (const PortableServer_POA src,
+			 PortableServer_POA dest)
+{
+	dest->p_thread              = src->p_thread;
+	dest->p_lifespan            = src->p_lifespan;
+	dest->p_id_uniqueness       = src->p_id_uniqueness;
+	dest->p_id_assignment       = src->p_id_assignment;
+	dest->p_servant_retention   = src->p_servant_retention;
+	dest->p_request_processing  = src->p_request_processing;
+	dest->p_implicit_activation = src->p_implicit_activation;
+}
+
 static void
 ORBit_POA_set_policies (PortableServer_POA      poa,
 			const CORBA_PolicyList *policies,
@@ -635,13 +712,41 @@ ORBit_POA_new (CORBA_ORB                  orb,
 			(GHashFunc) ORBit_sequence_CORBA_octet_hash,
 			(GEqualFunc) ORBit_sequence_CORBA_octet_equal);
 
-	poa->base.lock = linc_mutex_new ();
+	poa->base.lock = link_mutex_new ();
 
 	ORBit_POAManager_register_poa (manager, poa);
 
 	return ORBit_RootObject_duplicate (poa);
 }
 
+PortableServer_POA
+ORBit_POA_new_from (CORBA_ORB                  orb,
+		    PortableServer_POA         parent,
+		    const CORBA_char          *adaptor_name,
+		    const CORBA_PolicyList    *opt_policies,
+		    CORBA_Environment         *ev)
+{
+	PortableServer_POA poa;
+	
+	g_return_val_if_fail (parent != CORBA_OBJECT_NIL, CORBA_OBJECT_NIL);
+
+	poa = ORBit_POA_new (orb, adaptor_name, parent->poa_manager, NULL, ev);
+
+	g_return_val_if_fail (poa != CORBA_OBJECT_NIL, CORBA_OBJECT_NIL);
+
+	ORBit_POA_copy_policies (parent, poa);
+
+	if (opt_policies) {
+		int i;
+		for (i = 0; i < opt_policies->_length; i++)
+			ORBit_POA_set_policy (poa, opt_policies->_buffer[i]);
+	}
+
+	ORBit_POA_add_child (parent, poa);
+
+	return poa;
+}
+
 static CORBA_Object
 ORBit_POA_obj_to_ref (PortableServer_POA  poa,
 		      ORBit_POAObject     pobj,
@@ -708,6 +813,8 @@ ORBit_POAObject_release_cb (ORBit_RootObject robj)
 	/* object *must* be deactivated */
 	g_assert (pobj->servant == NULL);
 
+	giop_thread_key_release (robj);
+
 	object_id = pobj->object_id;
 	pobj->object_id = NULL;
 
@@ -812,7 +919,6 @@ ORBit_POA_activate_object_T (PortableServer_POA          poa,
 	g_assert (pobj->use_cnt == 0);
 
 	class_info = ORBIT_SERVANT_TO_CLASSINFO (servant);
-	g_assert (class_info->vepvmap);
 	pobj->vepvmap_cache = class_info->vepvmap;
 
 	pobj->servant = servant;
@@ -1079,10 +1185,10 @@ ORBit_POAObject_handle_request (ORBit_POAObject    pobj,
 	}
 
 	pobj->use_cnt++;
-	LINC_MUTEX_LOCK (poa->orb->lock);
+	LINK_MUTEX_LOCK (poa->orb->lock);
 	poa->orb->current_invocations =
 		g_slist_prepend (poa->orb->current_invocations, pobj);
-	LINC_MUTEX_UNLOCK (poa->orb->lock);
+	LINK_MUTEX_UNLOCK (poa->orb->lock);
 	
 	if (ev->_major == CORBA_NO_EXCEPTION && !pobj->servant)
 		CORBA_exception_set_system (
@@ -1164,11 +1270,11 @@ ORBit_POAObject_handle_request (ORBit_POAObject    pobj,
 			break;
 		}
 
-	LINC_MUTEX_LOCK (poa->orb->lock);
-	g_assert ((ORBit_POAObject)poa->orb->current_invocations->data == pobj);
+	LINK_MUTEX_LOCK (poa->orb->lock);
+/*	g_assert ((ORBit_POAObject)poa->orb->current_invocations->data == pobj); */
 	poa->orb->current_invocations =
 		g_slist_remove (poa->orb->current_invocations, pobj);
-	LINC_MUTEX_UNLOCK (poa->orb->lock);
+	LINK_MUTEX_UNLOCK (poa->orb->lock);
 	pobj->use_cnt--;
 
 	if (pobj->life_flags & ORBit_LifeF_NeedPostInvoke)
@@ -1235,6 +1341,16 @@ push_request_T (GIOPThread       *thread,
 		(thread, (gpointer *)pobj, (gpointer *)recv_buffer);
 }
 
+static void
+pool_push_request_for_T (gpointer          key_object,
+			 ORBit_POAObject  *pobj,
+			 GIOPRecvBuffer  **recv_buffer)
+{
+	g_assert (giop_thread_io ());
+	giop_thread_request_push_key
+		(key_object, (gpointer *)pobj, (gpointer *)recv_buffer);
+}			 
+
 typedef struct {
 	ORBit_POAObject  pobj;
 	GIOPRecvBuffer  *recv_buffer;
@@ -1336,7 +1452,7 @@ ORBit_POA_handle_request (PortableServer_POA poa,
 	else {
 		switch (poa->p_thread) {
 			case PortableServer_SINGLE_THREAD_MODEL:
-				if (giop_threaded ())
+				if (giop_thread_io ())
 					push_request_T (giop_thread_get_main (),
 							&pobj, &recv_buffer);
 				break;
@@ -1344,6 +1460,23 @@ ORBit_POA_handle_request (PortableServer_POA poa,
 				ORBit_ObjectAdaptor adaptor = (ORBit_ObjectAdaptor) poa;
 				
 				switch (adaptor->thread_hint) {
+				case ORBIT_THREAD_HINT_PER_OBJECT:
+					pool_push_request_for_T (pobj, &pobj, &recv_buffer);
+					break;
+
+				case ORBIT_THREAD_HINT_PER_POA:
+					pool_push_request_for_T (poa, &pobj, &recv_buffer);
+					break;
+
+				case ORBIT_THREAD_HINT_PER_CONNECTION:
+					pool_push_request_for_T (recv_buffer->connection,
+								 &pobj, &recv_buffer);
+					break;
+
+				case ORBIT_THREAD_HINT_PER_REQUEST:
+					pool_push_request_for_T (NULL, &pobj, &recv_buffer);
+					break;
+
 				case ORBIT_THREAD_HINT_ONEWAY_AT_IDLE:
 					if (!poa_recv_is_oneway (pobj, recv_buffer))
 						push_request_T (giop_thread_get_main (),
@@ -1352,8 +1485,9 @@ ORBit_POA_handle_request (PortableServer_POA poa,
 				case ORBIT_THREAD_HINT_ALL_AT_IDLE:
 					push_request_idle (&pobj, &recv_buffer);
 					break;
+
 				case ORBIT_THREAD_HINT_NONE:
-					if (giop_threaded ())
+					if (giop_thread_io ())
 						push_request_T (giop_thread_get_main (),
 								&pobj, &recv_buffer);
 					break;
@@ -1373,9 +1507,9 @@ ORBit_POA_handle_request (PortableServer_POA poa,
 	}
 	
  send_sys_ex:
-	ORBit_POAObject_invoke_incoming_request (pobj, recv_buffer, &env);
-
 	POA_UNLOCK (poa);
+
+	ORBit_POAObject_invoke_incoming_request (pobj, recv_buffer, &env);
 }
 
 static PortableServer_Servant
@@ -2014,14 +2148,14 @@ PortableServer_POA_servant_to_id (PortableServer_POA            poa,
 		 * The stricter form could be implemented, 
 		 * but it would only add more code...
 		 */
-		LINC_MUTEX_LOCK (poa->orb->lock);
+		LINK_MUTEX_LOCK (poa->orb->lock);
 		for (l = poa->orb->current_invocations; l; l = l->next) {
 			ORBit_POAObject pobj = l->data;
 
 			if (pobj->servant == servant)
 				objid = ORBit_sequence_CORBA_octet_dup (pobj->object_id);
 		}
-		LINC_MUTEX_UNLOCK (poa->orb->lock);
+		LINK_MUTEX_UNLOCK (poa->orb->lock);
 	}
 
 	if (!objid)
@@ -2079,14 +2213,14 @@ PortableServer_POA_servant_to_reference (PortableServer_POA            poa,
 		 * go backward from servant to pobj to use_cnt, but we
 		 * dont do this since forward search is more general 
 		 */
-		LINC_MUTEX_LOCK (poa->orb->lock);
+		LINK_MUTEX_LOCK (poa->orb->lock);
 		for (l = poa->orb->current_invocations; l; l = l->next) {
 			ORBit_POAObject pobj = l->data;
 
 			if (pobj->servant == servant)
 				result = ORBit_POA_obj_to_ref (poa, pobj, NULL, ev);
 		}
-		LINC_MUTEX_UNLOCK (poa->orb->lock);
+		LINK_MUTEX_UNLOCK (poa->orb->lock);
 	}
 
 	if (result == CORBA_OBJECT_NIL)
@@ -2227,6 +2361,139 @@ PortableServer_POA_id_to_reference (PortableServer_POA             poa,
 void
 ORBit_poa_init (void)
 {
-	_ORBit_poa_manager_lock = linc_mutex_new ();
+	ORBit_class_assignment_lock = link_mutex_new ();
+	_ORBit_poa_manager_lock = link_mutex_new ();
 	giop_thread_set_main_handler (ORBit_POAObject_invoke_incoming_request);
 }
+
+gboolean
+ORBit_poa_allow_cross_thread_call (ORBit_POAObject pobj)
+{
+	gpointer key = NULL;
+	GIOPThread *self;
+	PortableServer_POA poa = pobj->poa;
+
+	if (!poa)
+		return TRUE;
+
+	self = giop_thread_self ();
+
+	switch (poa->p_thread) {
+	case PortableServer_SINGLE_THREAD_MODEL:
+		break;
+
+	case PortableServer_ORB_CTRL_MODEL: {
+		ORBit_ObjectAdaptor adaptor = (ORBit_ObjectAdaptor) poa;
+				
+		switch (adaptor->thread_hint) {
+		case ORBIT_THREAD_HINT_PER_OBJECT:
+			key = pobj;
+			break;
+			
+		case ORBIT_THREAD_HINT_PER_POA:
+			key = poa;
+			break;
+			
+		case ORBIT_THREAD_HINT_PER_CONNECTION:
+			/* FIXME: ? */
+		case ORBIT_THREAD_HINT_PER_REQUEST:
+			return TRUE;
+			break;
+			
+		case ORBIT_THREAD_HINT_ONEWAY_AT_IDLE:
+		case ORBIT_THREAD_HINT_ALL_AT_IDLE:
+		case ORBIT_THREAD_HINT_NONE:
+			break;
+		}
+		break;
+	}
+	}
+
+	giop_thread_new_check (self);
+
+	if (!key)
+		return (self == giop_thread_get_main ());
+	else
+		return giop_thread_same_key (key, TRUE);
+}
+
+static gpointer
+get_c_method (CORBA_Object                 obj,
+	      glong                        class_id,
+	      PortableServer_ServantBase **servant,
+	      glong                        method_offset)
+{
+	guchar *epv_start;
+	ORBit_POAObject pobj;
+
+	if (!obj ||
+	    !(pobj = (ORBit_POAObject) obj->adaptor_obj) ||
+	    !(*servant = (PortableServer_ServantBase *) pobj->servant))
+		return NULL;
+
+	if (method_offset <= 0 || class_id <= 0)
+		return NULL;
+
+	if (!ORBit_poa_allow_cross_thread_call (pobj))
+		return NULL;
+
+	if (ORBit_small_flags & ORBIT_SMALL_FORCE_GENERIC_MARSHAL)
+		return NULL;
+
+	/*
+	 * FIXME: we could propagate the size of the vepvmap_cache
+	 * to here, and check it to avoid some really bogus stuff.
+	 */
+	if (!class_id || !(pobj->base.interface->adaptor_type & ORBIT_ADAPTOR_POA) ||
+	    !pobj->vepvmap_cache)
+		return NULL;
+
+	epv_start = (guchar *) (*servant)->vepv [ pobj->vepvmap_cache [class_id] ];
+
+	if (!epv_start)
+		return NULL;
+
+	return *(gpointer *)(epv_start + method_offset);
+}
+
+void
+ORBit_c_stub_invoke (CORBA_Object        obj,
+		     ORBit_IMethods     *methods,
+		     glong               method_index,
+		     gpointer            ret,
+		     gpointer            args,
+		     CORBA_Context       ctx,
+		     CORBA_Environment  *ev,
+
+		     glong               class_id,
+		     glong               method_offset,
+		     ORBitSmallSkeleton  skel_impl)
+{
+	gpointer method_impl;
+	PortableServer_ServantBase *servant;
+
+	if (skel_impl &&
+	    (method_impl = get_c_method (obj, class_id,
+					 &servant, method_offset))) {
+		
+		/* Unwound PreCall
+		   ++( ((ORBit_POAObject)(obj)->adaptor_obj)->use_cnt );	    \
+		   (obj)->orb->current_invocations =                           \
+		   g_slist_prepend ((obj)->orb->current_invocations,   \
+		   (obj)->adaptor_obj);               \
+		*/
+
+		skel_impl (servant, ret, args, ctx, ev, method_impl);
+
+		/* Unwound PostCall
+		   (obj)->orb->current_invocations =                                                      \
+		   g_slist_remove ((obj)->orb->current_invocations, pobj);                        \
+		   --(((ORBit_POAObject)(obj)->adaptor_obj)->use_cnt);                                    \
+		   if (((ORBit_POAObject)(obj)->adaptor_obj)->life_flags & ORBit_LifeF_NeedPostInvoke)    \
+		   ORBit_POAObject_post_invoke (((ORBit_POAObject)(obj)->adaptor_obj));           \
+		*/
+	} else
+		ORBit_small_invoke_stub_n
+			(obj, methods, method_index,
+			 ret, args, ctx, ev);
+}
